import pandas as pd
import numpy as np

from tensorflow.keras import Sequential,layers,losses,utils,Input
from tensorflow.keras.layers import Dense,LSTM,Dropout,concatenate,Flatten, Conv1D, MaxPooling1D,Activation,RepeatVector,TimeDistributed
import tensorflow as tf
# from tensorflow.keras.callbacks import ModelCheckpoint,TensorBoard
from tensorflow.keras.models import Model
# from tensorflow.keras.layers import Bidirectional,GRU,Lambda,Dot,Concatenate
from sklearn.metrics import mean_absolute_error,r2_score,mean_squared_error,mean_absolute_percentage_error
from keras.layers import Bidirectional,LSTM,GRU,Lambda,Dot,Concatenate

import matplotlib.pyplot as plt

from keras import optimizers
from keras.models import load_model

# 读取训练集
dushendux_train=pd.read_csv('shendux_train.csv')
dushendux_train1=np.array(dushendux_train)
dushendux_train2=np.reshape(dushendux_train1,(-1,120,7))

dushenduy_train=pd.read_csv('shenduy_train.csv')
dushenduy_train2=np.array(dushenduy_train)
# 读取测试集
dushendux_test=pd.read_csv('shendux_test.csv')
dushendux_test1=np.array(dushendux_test)
dushendux_test2=np.reshape(dushendux_test1,(-1,120,7))

dushenduy_test=pd.read_csv('shenduy_test.csv')
dushenduy_test2=np.array(dushenduy_test)

# # # 构造批数据
def create_batch_dataset(x,y,train=True,buffer_size=1000,batch_size=128):#buffer_size=1000表示可以打乱窗口里面的数据
    batch_data=tf.data.Dataset.from_tensor_slices((tf.constant(x),tf.constant(y)))#数据封装，tensor类型
    if train:
        return batch_data.cache().shuffle(buffer_size).batch(batch_size)
    else:
        return batch_data.batch(batch_size)

train_data_single=create_batch_dataset(dushendux_train2, dushenduy_train2,train=True)
val_data_single=create_batch_dataset(dushendux_test2,dushenduy_test2,train=False)

# 添加注意力机制
def attention_bilstm(inputs):
    hidden_states = inputs
    hidden_size = int(hidden_states.shape[2])
    score_first_part = Dense(hidden_size, use_bias=False, name='attention_score_vec')(hidden_states)
    h_t = Lambda(lambda x: x[:, -1, :], output_shape=(hidden_size,), name='last_hidden_state')(hidden_states)
    score = Dot(axes=[1, 2], name='attention_score')([h_t, score_first_part])
    attention_weights = Activation('softmax', name='attention_weight')(score)
    context_vector = Dot(axes=[1, 1], name='context_vector')([hidden_states, attention_weights])
    pre_activation = Concatenate(name='attention_output')([context_vector, h_t])
    attention_vector = Dense(32, use_bias=False, activation='tanh', name='attention_vector')(pre_activation)
    return attention_vector

time_steps=120
input_dim=8
model_input=Input(shape=(time_steps,input_dim))
num1_lstm2=LSTM(128,dropout=0.1,return_sequences=True)(model_input)
num2_lstm2=LSTM(64,dropout=0.1,return_sequences=True)(num1_lstm2)
attlstm2=attention_bilstm(num2_lstm2)
tlstm2=Dense(1)(attlstm2)
attlstm_model2=Model(model_input,tlstm2)
adaa=optimizers.Adam(lr=0.0005)
attlstm_model2.compile(optimizer=adaa, loss='mse',metrics='mae')# metrics=['accuracy']
attlstm_history2 = attlstm_model2.fit(train_data_single, validation_data=val_data_single,epochs=1400, verbose=1)
# 预测
attlstm2=attlstm_model2.predict(dushendux_test2,verbose=1)
attlstm2_modify = attlstm2*(attlstm2>=0)
zuiattlstmr2=r2_score(dushenduy_test2,attlstm2_modify)

maxmin=[662.0,3.0]
labels= np.array(dushenduy_test2)
for k in range(labels.shape[1]):
    labels[:, k] = labels[:, k] * (maxmin[0] - maxmin[1]) + maxmin[1]  # 反归一化
# 反归一化
attlstm2_modify= np.array(attlstm2_modify)  # 转换为numpy形式
attlstm2_modify= attlstm2_modify.reshape((-1,1))
attlstm2_modify_len= attlstm2_modify.shape[1]
for k in range(attlstm2_modify_len):
    attlstm2_modify[:, k] = attlstm2_modify[:, k] * (maxmin[0] - maxmin[1]) + maxmin[1]  # 反归一化

pd.DataFrame(attlstm_history2.history).to_csv('LSTM-Att_history.csv', index=False)
lstmatt_data=pd.DataFrame(attlstm2_modify,columns=['att-lstm_pred'])
lstmatt_data.to_csv('LSTM-Att_pred.csv',index=False)

lstmatt_model_history=pd.read_csv('LSTM-Att_history.csv')
lstmatt_pre=pd.read_csv('LSTM-Att_pred.csv')
lstm_rmse=np.sqrt(mean_squared_error(labels,lstmatt_pre))
lstm_mae=mean_absolute_error(labels,lstmatt_pre)

# 画图
plt.figure(figsize=(12,6))
plt.rcParams['font.sans-serif']=['Fangsong']
plt.rcParams['axes.unicode_minus']=False
plt.grid(color="k", linestyle=":")
plt.plot(lstmatt_model_history['val_loss'],label=['LSTM-Attention_loss_value'],color='k')
plt.xlabel('number',fontsize=25)
plt.ylabel('LSTM-Attention_loss_value',fontsize=25)
plt.xticks(fontsize=20)
plt.yticks(fontsize=20)
plt.rcParams['xtick.direction'] = 'in'
plt.rcParams['ytick.direction'] = 'in'
plt.legend(fontsize=20)
plt.savefig("LSTM-Attention损失图.jpg")
plt.show()

plt.figure(figsize=(18, 6))
ax = plt.subplot()
plt.rcParams['font.sans-serif']=['FangSong']
plt.rcParams['axes.unicode_minus']=False
plt.grid(color="k", linestyle=":")
ax.set_title('LSTM-Attention model',fontsize=25)
x = range(0,500,1)
plt.plot(x,labels[0:500],"k:", label="real", alpha=0.8,ms=10)
plt.plot(x,lstmatt_pre[0:500], "k-", label="pred", alpha=0.9,ms=10)
plt.xlabel('time/h',fontsize=25)
plt.ylabel('PM2.5/(ug/m^3)',fontsize=25)
plt.tick_params(labelsize=20)
plt.legend(loc='upper right',fontsize=20)
plt.savefig("LSTM-Attention拟合图500.jpg")